/* hub.c -- implementazione hub: protocollo NMDC */
/******************************************************************************
* DCLibC - Direct Connect C library                                           *
*                                                                             *
* Copyright (C) 2005-2007 daniele_athome                                      *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; either version 2              *
* of the License, or (at your option) any later version.                      *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
*******************************************************************************/

#include <string.h>
#include <sys/socket.h>
#include <errno.h>
#include <unistd.h>
#include <netdb.h>
#include "dclibc.h"
#include "hubconn.h"
#include "userinfo.h"
#include "userlist.h"
#include "hub.h"

static gboolean hub_closer(hub_t * hub);
static void hubconn_notify(HubConn * conn,HUBCONN_NOTIFY_TYPES notify_type,gpointer data);

static GPtrArray * hubs = NULL;

extern void stop_connecting(HubConn * conn);

/* crea una nuova struct hub_t */
hub_t * dc_hub_new(const char * server,port_t port,userinfo_t * userinfo,userlist_t * userlist,HubNotifyFunc notify_func,gpointer user_data)
{
	if(server == NULL || port < 1 || port > 65535 || userinfo == NULL || userlist == NULL) return NULL;
	hub_t * hub = g_malloc0(sizeof(hub_t));
	HubConn * cn = dc_hub_conn_new(server,port,hubconn_notify,hub);
	if(cn == NULL)
	{
		g_free(hub);
		return NULL;
	}
	hub->conn = cn;
	hub->hubserver = g_strdup(server);
	hub->hubport = port;
	hub->hubname = g_strdup(server);
	hub->hubtopic = NULL;
	hub->userinfo = userinfo;
	hub->userlist = userlist;
	hub->notify_func = notify_func;
	hub->user_data = user_data;
	hub->unparsed = g_string_new("");
	hub->parsetag = -1;
	hub->notify_queue = g_queue_new();
	if(hubs == NULL) hubs = g_ptr_array_new();
	g_ptr_array_add(hubs,hub);
	return hub;
}

/* distrugge una struct hub_t */
void dc_hub_delete(hub_t * hub)
{
	if(hub != NULL)
	{
		/* chiudi la connessione */
		g_debug("(dc_hub_delete): closing connection...");
		dc_hub_disconnect(hub);
		g_idle_add((GSourceFunc)hub_closer,hub);
	}
}

static gboolean hub_closer(hub_t * hub)
{
	/* distruggi tutto */
	while(g_source_remove_by_user_data(hub)) g_debug("Removing another one...");
	while(!g_queue_is_empty(hub->notify_queue))
	{
		g_debug("Removing another two...");
		g_source_remove(GPOINTER_TO_INT(g_queue_pop_head(hub->notify_queue)));
	}
	if(hub->parsetag >= 0) g_source_remove(hub->parsetag);
	dc_hub_conn_delete(hub->conn);
	g_free(hub->hubserver);
	g_free(hub->hubname);
	g_free(hub->sr_host);
	g_string_free(hub->unparsed,TRUE);
	g_debug("closer: hubs is %p",hubs);
	if(hubs != NULL) g_ptr_array_remove(hubs,hub);
	g_free(hub);
	return FALSE;
}

/* si connette ad un hub e avvia la procedura di login sincrona */
gboolean dc_hub_connect(hub_t * hub)
{
	if(hub == NULL) return FALSE;
	hub->handshake = TRUE;
	dc_hub_conn_open(hub->conn);
	return TRUE;
}

/* disconnette un hub */
void dc_hub_disconnect(hub_t * hub)
{
	if(hub != NULL) dc_hub_conn_close(hub->conn);
}

/* disconnette tutti gli hub */
void _dc_hub_disconnect_all(void)
{
	if(hubs != NULL)
	{
		hub_t * hb = NULL;
		int i;
		for(i=0;i<hubs->len;i++)
		{
			hb = g_ptr_array_index(hubs,i);
			//g_debug("(disconnect_all): disconnecting %p",hb);
			if(hb != NULL)
			{
				stop_connecting(hb->conn);
				dc_hub_delete(hb);
			}
		}
		g_ptr_array_free(hubs,TRUE);
		hubs = NULL;
	}
}

/* riceve le notifiche dalla connessione */
static void hubconn_notify(HubConn * conn,HUBCONN_NOTIFY_TYPES notify_type,gpointer data)
{
	g_debug("(hubconn_notify): conn = %p, notify = %d, data = %p",conn,notify_type,data);
	hub_t * hub = conn->user_data;
	if(hub == NULL) return;
	switch(notify_type)
	{
		case CONN_NOTIFY_SYSERROR:
			HUB_NOTIFY(hub,NOTIFY_SYSERROR,NULL,NULL);
			break;
		case CONN_NOTIFY_ERROR:
			HUB_NOTIFY(hub,NOTIFY_ERROR,NULL,NULL);
			break;
		case CONN_NOTIFY_CONNECTING:
			HUB_NOTIFY(hub,NOTIFY_CONNECTING,NULL,NULL);
			break;
		case CONN_NOTIFY_DISCONNECTED:
			HUB_NOTIFY(hub,NOTIFY_DISCONNECTED,NULL,NULL);
			break;
		case CONN_NOTIFY_ERROR_RESOLV:
			HUB_NOTIFY(hub,NOTIFY_ERROR_RESOLV,NULL,NULL);
			break;
		case CONN_NOTIFY_ERROR_CONNECT:
			HUB_NOTIFY(hub,NOTIFY_ERROR_CONNECT,NULL,NULL);
			break;
		case CONN_NOTIFY_ERROR_REFUSED:
			HUB_NOTIFY(hub,NOTIFY_ERROR_REFUSED,NULL,NULL);
			break;
		case CONN_NOTIFY_ERROR_TIMEOUT:
			HUB_NOTIFY(hub,NOTIFY_ERROR_TIMEOUT,NULL,NULL);
			break;
		case CONN_NOTIFY_RESOLVED:
			g_debug("(hubconn_notify): address resolved (%s)",conn->serverip);
			break;
		case CONN_NOTIFY_CONNECTED:
			HUB_NOTIFY(hub,NOTIFY_CONNECTED,NULL,NULL);
			break;
		case CONN_NOTIFY_DATA:
			g_debug("(hubconn_notify_data): conn = %p, notify = %d, data = %p",conn,notify_type,data);
			if(hub->parsetag < 0) hub->parsetag = g_idle_add((GSourceFunc)dc_proto_parse,hub);
			/*
			if(hub->unparsed->len > 0)
			{
				g_warning("(process_data): unparsed: \"%s\"",hub->unparsed->str);
				g_string_erase(hub->unparsed,0,-1);
			}
			*/
			break;
		case CONN_NOTIFY_NONE:
		default:
			g_warning("(hubconn_notify): unknown notify %d",notify_type);
	}
}

typedef struct
{
	hub_t * hub;
	gint notify_type;
	gpointer data;
	GDestroyNotify destroy_data;
} internal_notify;

static gboolean call_notifier(internal_notify * n)
{
	if(n != NULL)
	{
		g_queue_pop_head(n->hub->notify_queue);
		((HubNotifyFunc)n->hub->notify_func)(n->hub,n->notify_type,n->data);
		if(n->destroy_data != NULL) (n->destroy_data)(n->data);
	}
	return FALSE;
}

void HUB_NOTIFY(hub_t * hub,gint notify_type,gpointer data,GDestroyNotify destroy_data)
{
	if(hub != NULL)
	{
		internal_notify * n = g_malloc0(sizeof(internal_notify));
		n->hub = hub;
		n->notify_type = notify_type;
		n->data = data;
		n->destroy_data = destroy_data;
		g_queue_push_tail(hub->notify_queue,GINT_TO_POINTER(g_idle_add_full(G_PRIORITY_DEFAULT_IDLE,(GSourceFunc)call_notifier,n,g_free)));
	}
}
