/* hubconn.c -- gestisce la connessione ad un hub */
/******************************************************************************
* DCLibC - Direct Connect C library                                           *
*                                                                             *
* Copyright (C) 2005-2007 daniele_athome                                      *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; either version 2              *
* of the License, or (at your option) any later version.                      *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
*******************************************************************************/

#include <glib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <signal.h>
#include <netdb.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include "hubconn.h"
#include "dcproto.h"
#include "hub.h"

static gboolean connector(HubConn * conn);
static gboolean disconnector(HubConn * conn);
static gboolean freedom(HubConn * conn);
static gboolean do_connect(HubConn * conn);
static void connector_child(HubConn * conn);
static gboolean connector_read_child(GIOChannel *source,GIOCondition condition,gpointer data);
void stop_connecting(HubConn * conn);
static gboolean connection_timeout(HubConn * conn);
static char * extract_child(const char * line);
static void start_connected(HubConn * conn);
static gboolean process_data(GIOChannel *source,GIOCondition condition,gpointer data);
static gboolean flush_data(HubConn * conn);

/* inizializza hubconn */
void dc_hub_conn_init(void)
{
	g_debug("(dc_hub_conn_init): initializing hub connection subsystem...");
	/* NOTHING TODO */
}

/* crea una nuova connessione ad un hub
 cb_notify: funzione che ricevera` le notifiche */
HubConn * dc_hub_conn_new(const char * server,port_t port,HubConnNotifier cb_notify,gpointer user_data)
{
	if(server == NULL || port < 1 || port > 65536) return NULL;
	HubConn * c = g_malloc0(sizeof(HubConn));
	c->sock_fd = -1;
	c->server = g_strdup(server);
	c->serverip = NULL;
	c->port = port;
	c->pid = -1;
	c->pipes[0] = c->pipes[1] = -1;
	c->timetag = c->iotag = c->socktag = c->outtag = -1;
	c->notifier = cb_notify;
	c->buffer = g_string_new("");
	c->out_queue = g_string_new("");
	c->user_data = user_data;
	c->connected = FALSE;
	return c;
}

void dc_hub_conn_delete(HubConn * conn)
{
	if(conn != NULL)
	{
		dc_hub_conn_close(conn);
		g_idle_add((GSourceFunc)freedom,conn);
	}
}

/* apre la connessione ad un hub */
void dc_hub_conn_open(HubConn * conn)
{
	if(conn != NULL) g_idle_add((GSourceFunc)connector,conn);
}

// decommentare la riga sotto assegnare una priorita` alta all'evento di disconnessione
//#define DISCONNECT_PRIORITY
/* chiude la connessione ad un hub e libera tutte le risorse occupate, incluso <conn> */
void dc_hub_conn_close(HubConn * conn)
{
	if(conn != NULL)
	{
		g_debug("connected = %s, queuing disconnector...",(conn->connected)?"TRUE":"FALSE");
		if(conn->connected)
#ifdef DISCONNECT_PRIORITY
		 g_idle_add_full(G_PRIORITY_HIGH_IDLE,(GSourceFunc)disconnector,conn,NULL);
#else
		 g_idle_add((GSourceFunc)disconnector,conn);
#endif
	}
}

static gboolean freedom(HubConn * conn)
{
	g_free(conn->server);
	g_free(conn->serverip);
	g_string_free(conn->buffer,TRUE);
	g_string_free(conn->out_queue,TRUE);
	g_free(conn);
	return FALSE;
}

static gboolean disconnector(HubConn * conn)
{
	stop_connecting(conn);
	if(conn->socktag >= 0) g_source_remove(conn->socktag);
	conn->socktag = -1;
	if(conn->outtag >= 0) g_source_remove(conn->outtag);
	conn->outtag = -1;
	g_debug("(disconnector): closing socket %d",conn->sock_fd);
	if(conn->sock_fd >= 0) close(conn->sock_fd);
	conn->sock_fd = -1;
	return FALSE;
}

void stop_connecting(HubConn * conn)
{
	if(conn != NULL)
	{
		g_debug("(stop_connecting): killing pid %d",conn->pid);
		if(conn->pipes[0] >= 0) close(conn->pipes[0]);
		if(conn->pipes[1] >= 0) close(conn->pipes[1]);
		if(conn->pid >= 0) kill(conn->pid,SIGKILL);
		conn->pid = -1;
		g_source_remove(conn->iotag);
		g_source_remove(conn->timetag);
		conn->iotag = conn->timetag = -1;
		conn->connected = FALSE;
	}
}

static gboolean connector(HubConn * conn)
{
	if(!conn->connected) if(!do_connect(conn)) dc_hub_conn_close(conn);
	return FALSE;
}

static gboolean do_connect(HubConn * conn)
{
	HUBCONN_NOTIFY(conn,CONN_NOTIFY_CONNECTING,NULL);
	if(conn == NULL)
	{
		g_critical("(do_connect): connection is NULL");
		goto error1;
	}
	/* crea la pipe per comunicare con il figlio */
	if(pipe(conn->pipes) < 0)
	{
		g_critical("(do_connect): cannot create a pipe pair");
		goto error1;
	}
	/* creiamo la socket */
	conn->sock_fd = socket(PF_INET,SOCK_STREAM,0);
	if(conn->sock_fd < 0)
	{
		g_critical("(do_connect): cannot create socket");
		goto error2;
	}
	socklen_t sw;
	sw = 1;
	setsockopt(conn->sock_fd,SOL_SOCKET,SO_REUSEADDR,(char *)&sw,sizeof(sw));
	sw = 1;
	setsockopt(conn->sock_fd,SOL_SOCKET,SO_KEEPALIVE,(char *)&sw,sizeof(sw));
	/* forkettamo! */
	conn->pid = fork();
	if(conn->pid < 0)
	{
		g_critical("(do_connect): fork failed!");
		goto error3;
		return FALSE;
	}
	else if(!conn->pid)	// siamo nel figlio
	{
		connector_child(conn);
		_exit(EXIT_SUCCESS);
	}
	conn->connected = TRUE;
	// siamo nel padre, prepariamo il timeout e poniamo in ascolto la pipe per la lettura
	conn->timetag = g_timeout_add(60000,(GSourceFunc)connection_timeout,conn);
	GIOChannel * ioc = g_io_channel_unix_new(conn->pipes[0]);
	g_io_channel_set_encoding(ioc,NULL,NULL);
	conn->iotag = g_io_add_watch(ioc,G_IO_IN,connector_read_child,conn);
	return TRUE;
error3:
	close(conn->sock_fd);
	conn->sock_fd = -1;
error2:
	close(conn->pipes[0]);
	close(conn->pipes[1]);
	conn->pipes[0] = conn->pipes[1] = -1;
error1:
	return FALSE;
}

static gboolean connection_timeout(HubConn * conn)
{
	dc_hub_conn_close(conn);
	HUBCONN_NOTIFY(conn,CONN_NOTIFY_ERROR_TIMEOUT,NULL);
	return FALSE;
}

/* processo figlio che sta provando a connettersi */
static void connector_child(HubConn * conn)
{
	/* risolviamo l'host prima di tutto */
	struct addrinfo * res = NULL;
	int n = getaddrinfo(conn->server,NULL,NULL,&res);
	char * cc = NULL;
	if(n || res == NULL)
	{
		g_critical("(resolver_thread): Error resolving host");
		write(conn->pipes[1],"1\n",2); // errore risoluzione
		return;
	}
	// invia il primo ip al processo padre
	struct in_addr ss;
	struct in_addr * host = g_malloc0(sizeof(struct in_addr));
	struct sockaddr_in * sockin = (struct sockaddr_in *)res->ai_addr;
	memcpy(host,&sockin->sin_addr,sizeof(struct in_addr));
	memcpy(&ss,&sockin->sin_addr,sizeof(struct in_addr));
	//g_debug("(resolver_thread): copying \"%s\" for \"%s\"",inet_ntoa(ss),socket->server);
	cc = g_strdup_printf("2%s\n",inet_ntoa(ss));	// risolto!
	write(conn->pipes[1],cc,strlen(cc));
	g_free(cc);
	/* ok, ora proviamo a connetterci */
	struct sockaddr_in serv_addr;
	memset((void *)&serv_addr,(char)0,sizeof(struct sockaddr_in));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(conn->port);
	memcpy((void *)&serv_addr.sin_addr,(void *)host,sizeof(struct in_addr));
	g_free(host);
	if(connect(conn->sock_fd,(struct sockaddr *)&serv_addr,sizeof(struct sockaddr_in)) < 0)
	{
		/* uffa niente connessione ;) */
		if(errno == ECONNREFUSED) cc = "3\n";
		else cc = "4\n";
		write(conn->pipes[1],cc,strlen(cc));
		return;
	}
	int flags = fcntl(conn->sock_fd, F_GETFL, 0);
	if(flags < 0)
	{
		g_warning("(do_connect): fcntl(get) failed");
		write(conn->pipes[1],"0\n",2);
		return;
	}
	/* setta la connessione non bloccante */
	if(fcntl(conn->sock_fd, F_SETFL, flags | O_NONBLOCK) < 0)
	{
		g_warning("(do_connect): fcntl(set) failed");
		write(conn->pipes[1],"0\n",2);
		return;
	}
	/* ok, siamo connessi, inviamo la socket al processo padre */
	cc = g_strdup_printf("5%d\n",conn->sock_fd); // connesso!
	write(conn->pipes[1],cc,strlen(cc));
	g_free(cc);
}

/* interpreta i dati arrivati dal processo figlio che sta provando a connettersi */
static gboolean connector_read_child(GIOChannel *source,GIOCondition condition,gpointer data)
{
/*
stati dal processo figlio:
0: errore di sistema
1: errore risoluzione
2%s: risoluzione ok (ip in dot-number notation)
3: connessione rifiutata
4: errore di connessione
5%d: connessione ok (descrittore socket)
*/
	char * buf = NULL;
	gsize len = 0;
	GError * error = NULL;
	HubConn * conn = data;
	if(conn == NULL) return FALSE;	// possiamo anche andare via subito, tanto...
	g_debug("CONNECTOR: reading line (%d)...",condition);
	GIOStatus ret = g_io_channel_read_line(source,&buf,&len,NULL,&error);
	g_debug("CONNECTOR: again = %d, ret = %d",G_IO_STATUS_AGAIN,ret);
	if(ret == G_IO_STATUS_AGAIN) return TRUE; // riproviamo...
	else if((ret == G_IO_STATUS_NORMAL || ret == G_IO_STATUS_EOF) && (len > 0))
	{
		g_debug("CONNECTOR: normal = %d, eof = %d, ret = %d",G_IO_STATUS_NORMAL,G_IO_STATUS_EOF,ret);
		g_debug("(connector_read_child): \"%s\" (%c)",buf,buf[0]);
		switch(buf[0])
		{
			case '0':	// errore di sistema
				dc_hub_conn_close(conn);
				HUBCONN_NOTIFY(conn,CONN_NOTIFY_SYSERROR,NULL);
				break;
			case '1':	// errore risoluzione
				dc_hub_conn_close(conn);
				// notifica l'hub
				HUBCONN_NOTIFY(conn,CONN_NOTIFY_ERROR_RESOLV,NULL);
				break;
			case '2':	// indirizzo risolto
			{
				// estrai indirizzo
				conn->serverip = extract_child(buf);
				g_debug("CONNECTOR: addr: \"%s\"",conn->serverip);
				HUBCONN_NOTIFY(conn,CONN_NOTIFY_RESOLVED,NULL);
				break;
			}
			case '3':	// connessione rifiutata
				dc_hub_conn_close(conn);
				HUBCONN_NOTIFY(conn,CONN_NOTIFY_ERROR_REFUSED,NULL);
				break;
			case '4':	// errore di connessione
				dc_hub_conn_close(conn);
				HUBCONN_NOTIFY(conn,CONN_NOTIFY_ERROR_CONNECT,NULL);
				break;
			case '5':	// connesso
			{
				char * sk = extract_child(buf);
				g_debug("CONNECTOR: socket: \"%s\"",sk);
				int sfd = atoi(sk);
				g_free(sk);
				g_debug("CONNECTOR: descriptor: %d (0x%x)",conn->sock_fd,conn->sock_fd);
				if(conn->sock_fd != sfd)
				{
					dc_hub_conn_close(conn);
					HUBCONN_NOTIFY(conn,CONN_NOTIFY_SYSERROR,NULL);
				}
				else
				{
					stop_connecting(conn);
					conn->connected = TRUE;	// ripristina il connected
					HUBCONN_NOTIFY(conn,CONN_NOTIFY_CONNECTED,NULL);
					start_connected(conn);
				}
				break;
			}
			default:
				g_debug("CONNECTOR: unknown response (buf[0] = '%c')",buf[0]);
		}
		if(ret == G_IO_STATUS_EOF) return FALSE;
	}
	return TRUE;
}

/* richiamata appena dopo la connessione */
static void start_connected(HubConn * conn)
{
	if(conn != NULL)
	{
		if(conn->sock_fd < 0) return;
		GIOChannel * ioc = g_io_channel_unix_new(conn->sock_fd);
		g_io_channel_set_encoding(ioc,NULL,NULL);
		conn->socktag = g_io_add_watch(ioc,G_IO_IN|G_IO_PRI|G_IO_HUP|G_IO_ERR|G_IO_NVAL,process_data,conn);
		g_io_channel_unref(ioc);
	}
}

/* estrae e restituisce una stringa con il contenuto della linea <line> */
static char * extract_child(const char * line)
{
	char * pos = g_strdup(line+1);
	char * endl = strstr(pos,"\n");
	if(endl != NULL)
	{
		*endl = '\0';
		return pos;
	}
	return NULL;
}

/* notificatore di tutti gli eventi sulla socket */
static gboolean process_data(GIOChannel *source,GIOCondition condition,gpointer data)
{
	//g_debug("IOChannel event! condition = 0x%x, data = %p",condition,data);
	//g_debug("IN: 0x%x, OUT: 0x%x, PRI: 0x%x, ERR: 0x%x, HUP: 0x%x, NVAL: 0x%x",G_IO_IN,G_IO_OUT,G_IO_PRI,G_IO_ERR,G_IO_HUP,G_IO_NVAL);
	if(data == NULL) return FALSE;	// LOOL :D
	HubConn *conn = data;
	if(condition & G_IO_HUP || condition & G_IO_NVAL || condition & G_IO_ERR)
	{
		g_debug("something went wrong, disconnecting...");
		dc_hub_conn_close(conn);
		HUBCONN_NOTIFY(conn,CONN_NOTIFY_DISCONNECTED,NULL);
		return FALSE;
	}
	if(condition & G_IO_IN || condition & G_IO_PRI)
	{
		const int BUF_SIZE = 1024;
		char * buf = g_malloc0(BUF_SIZE+1);
		int ret = recv(conn->sock_fd,buf,BUF_SIZE,0);
		if(!ret)	// connessione chiusa
		{
			g_debug("connection reset, data = \"%s\"",buf);
			g_free(buf);
 			dc_hub_conn_close(conn);
 			HUBCONN_NOTIFY(conn,CONN_NOTIFY_DISCONNECTED,NULL);
			return FALSE;
		}
		else if(ret < 0 && errno != EAGAIN)	// errore
		{
			g_free(buf);
			dc_hub_conn_close(conn);
			HUBCONN_NOTIFY(conn,CONN_NOTIFY_ERROR_CONNECT,NULL);
			return FALSE;
		}
		//g_debug("Data: \"%s\"",buf);
		// appendi i dati al buffer di input
		g_string_append(conn->buffer,buf);
		g_free(buf);
		//g_debug("===== RECEIVED: \"%s\"",conn->buffer->str);
		#if 1
		if(conn->buffer->str[conn->buffer->len-1] == '|')	// abbiamo un fine comando, parsiamo
		#else
		//if(strstr(conn->buffer->str,"|"))
		#endif
		{
			HUBCONN_NOTIFY(conn,CONN_NOTIFY_DATA,NULL);
		}
	}
	return TRUE;
}

/* appende i dati alla coda dei dati da spedire */
void dc_hub_conn_append_data(HubConn * conn,char * buf)
{
	if(conn != NULL && buf != NULL)
	{
		g_string_append(conn->out_queue,buf);
		if(conn->outtag < 0) conn->outtag = g_idle_add((GSourceFunc)flush_data,conn);
	}
}

static gboolean flush_data(HubConn * conn)
{
	if(conn != NULL)
	{
		int i=0;
		g_debug("(flush_data): flushing: \"%s\"",conn->out_queue->str);
		size_t c = strlen(conn->out_queue->str);
		if(c > 0)
		{
			//g_debug("(pre) conn->sock_fd = %d",conn->sock_fd);
			if(conn->sock_fd >= 0)
			{
				struct timeval tv;
				tv.tv_sec = 0;
				tv.tv_usec = 0;
				fd_set writeset;
				FD_ZERO(&writeset);
				FD_SET(conn->sock_fd,&writeset);
				i = select(FD_SETSIZE,NULL,&writeset,NULL,&tv);
				FD_CLR(conn->sock_fd,&writeset);
				//g_debug("conn->sock_fd = %d, i = %d",conn->sock_fd,i);
				if(i > 0)
				{
					//g_debug("Sending...");
					send(conn->sock_fd,conn->out_queue->str,c,0);
					// svuota la coda di uscita
					//g_debug("Queueing off...");
					g_string_erase(conn->out_queue,0,-1);
					conn->outtag = -1;
				}
			}
		}
        }
	return FALSE;
}
