/* search.c -- sistema di ricerca */
/******************************************************************************
* DCLibC - Direct Connect C library                                           *
*                                                                             *
* Copyright (C) 2005-2007 daniele_athome                                      *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; either version 2              *
* of the License, or (at your option) any later version.                      *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
*******************************************************************************/

#include <string.h>
#include <glib.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <errno.h>
#include "hub.h"
#include "search.h"
#include "userinfo.h"
#include "dcproto.h"

#define SR_PACKET_SIZE		51200	// dimensione di un pacchetto di risultato (50 KB)

typedef struct
{
	hub_t * hub;		// hub sul quale abbiamo effettuato la ricerca
} internal_search_t;

typedef struct
{
	hub_t * hub;		// hub dal quale arriva il risultato (solo ricerca passiva)
	char * sr_packet;	// pacchetto UDP dal nick (ricerca attiva) / pacchetto TCP dall'hub (ricerca passiva)
	char mode;		// modalita` della ricerca ('A' o 'P')
} internal_result_t;

char ttbool[2] = { 'F', 'T' };
#define BOOL_TO_CHAR(b)		ttbool[b]

static GIOChannel * active_socket = NULL;
static gint active_watch = -1;

static gboolean sr_result(internal_result_t * res);
static void destroy_internal_result(internal_result_t * res);
void destroy_result(result_t * r);
void destroy_search(search_t * se);
static gboolean update_search_socket(hub_t * hub);
static gboolean active_arrival(GIOChannel *source,GIOCondition condition,hub_t * hub);

/* UDP SOCKET */
static GIOChannel * udp_socket_create(const char * ip,port_t port);
static int udp_socket_read(GIOChannel *ioc,char * buf,size_t len);

void dc_search_init(void)
{
	g_debug("(dc_search_init): initializing search subsystem...");
}

/* effettua una ricerca sull'hub <hub>
 size_restricted: indica se la ricerca e` ristretta alla dimensione
 is_max_size: indica se la dimensione indicata e` la massima, altrimenti la minima (valido se size_restricted == TRUE)
 size: indica la dimensione a cui e` limitata la ricerca (valido se size_restricted == TRUE)
 pattern: stringa di ricerca: <item1>$<item2>$<item3>$<item4>....$...<itemN>
 restituisce l'id della ricerca */
int dc_search_request(hub_t * hub,gboolean size_restricted,gboolean is_max_size,guint64 size,DATA_TYPE type,const char * pattern)
{
	if(hub == NULL || (size_restricted != FALSE && size_restricted != TRUE) || (is_max_size != FALSE && is_max_size != TRUE) || pattern == NULL) return -1;
	if(!hub->conn->connected) return -1;
	char mode = '\0';
	/* prepara il comando di ricerca */
	char * sr = NULL,*presr = NULL;
	g_debug("MODE: '%c'",userinfo_get_mode(hub->userinfo));
	presr = g_strdup_printf(DC_SEARCH_SYNTAX,BOOL_TO_CHAR(size_restricted),BOOL_TO_CHAR(is_max_size),size,type,pattern);
	mode = userinfo_get_mode(hub->userinfo);
	if(mode == TAG_MODE_ACTIVE)
	{
		if(hub->sr_host == NULL || hub->sr_port < 1 || hub->sr_port > 65536)
		{
			g_free(presr);
			return -1;
		}
		/* prepara la socket se necessario */
		if(active_watch < 0 || active_socket == NULL)
		{
			if(update_search_socket(hub) == FALSE) return -1;
		}
		sr = g_strdup_printf(DC_CMD_SEARCH(TAG_MODE_ACTIVE),hub->sr_host,hub->sr_port,presr);
	}
	else if(mode == TAG_MODE_PASSIVE)
	 sr = g_strdup_printf(DC_CMD_SEARCH(TAG_MODE_PASSIVE),hub->userinfo->nick,presr);
	else
	{
		g_free(presr);
		g_critical("(dc_search_request): unknown search mode");
		return -1;
	}
	g_free(presr);
	/* invia il comando di ricerca */
	g_debug("(dc_search_request): \"%s\"",sr);
	dc_proto_send(hub,sr);
	g_debug("(dc_search_request): search in progress...");
	g_free(sr);
	return 0;
}

/* ricrea la socket udp per la ricerca attiva */
static gboolean update_search_socket(hub_t * hub)
{
	if(hub != NULL)
	{
		g_debug("host = %p, port = %d",hub->sr_host,hub->sr_port);
		if(hub->sr_host != NULL && hub->sr_port >= 1 && hub->sr_port <= 65536)
		{
			g_debug("Active watch = %d",active_watch);
			if(active_watch >= 0) g_source_remove(active_watch);
			g_debug("Active socket = %p",active_socket);
			if(active_socket != NULL)
			{
				g_io_channel_shutdown(active_socket,FALSE,NULL);
				g_debug("Unreferencing...");
				g_io_channel_unref(active_socket);
			}
			g_debug("Creating active socket...");
			active_socket = udp_socket_create(/*hub->sr_host*/NULL,hub->sr_port);
			g_debug("Created! Active socket = %p",active_socket);
			if(active_socket != NULL)
			{
				g_debug("Adding watch...");
				active_watch = g_io_add_watch(active_socket,G_IO_IN|G_IO_PRI|G_IO_HUP|G_IO_ERR|G_IO_NVAL,
				 (GIOFunc)active_arrival,hub);
				g_debug("Active watch = %d",active_watch);
				//g_io_channel_unref(active_socket);
				if(active_watch < 0)
				{
					active_socket = NULL; // in caso di errore la socket e` distrutta
					return FALSE;
				}
				return TRUE;
			}
		}
	}
	return FALSE;
}

/* imposta i parametri per la socket di ascolto per la ricerca attiva
 restituisce TRUE se il bind e` avvenuto correttamente, altrimenti FALSE */
gboolean dc_search_set_active_search_params(hub_t * hub,const char * host,port_t port)
{
	if(hub != NULL || host != NULL || port >= 1 || port <= 65536)
	{
		g_debug("Settings hosts...");
		hub->sr_host = g_strdup(host);
		hub->sr_port = port;
		g_debug("Returning search socket...");
		return update_search_socket(hub);	// aggiorna la socket
	}
	return FALSE;
}

/* appende la notifica per il risultato
 hub: hub della ricerca
 sr_packet: pacchetto del responso */
void notify_result(hub_t * hub,char * sr_packet)
{
	g_debug("hub = %p, sr_packet = %p (\"%s\")",hub,sr_packet,sr_packet);
	if(hub == NULL || sr_packet == NULL) return;
	internal_result_t * res = g_malloc0(sizeof(internal_result_t));
	res->hub = hub;
	res->sr_packet = g_strdup(sr_packet);
	g_idle_add((GSourceFunc)sr_result,res);
}

/* gestisce un solo risultato e lo notifica al client
 data: internal_result_t */
static gboolean sr_result(internal_result_t * res)
{
	if(res != NULL)
	{
		g_warning("(sr_result): search result, data: \"%s\"",res->sr_packet);
		/* parsa il risultato */
		result_t * r = dc_proto_parse_sr(res->sr_packet);
		if(r != NULL)
		{
			/* notifica al client il risultato */
			HUB_NOTIFY(res->hub,NOTIFY_SEARCH_RESULT,r,(GDestroyNotify)destroy_result);
			/* libera il risultato */
		}
		destroy_internal_result(res);
	}
	return FALSE;
}

/* gestisce gli eventi sulla socket della ricerca attiva
 data: hub_t del primo hub dopo l'ultima modifica dei dati di socket che ha richiesto la ricerca
 ATTENZIONE: non contare sull'hub notificato come fonte della ricerca, potrebbe non essere quello giusto!! */
static gboolean active_arrival(GIOChannel *source,GIOCondition condition,hub_t * hub)
{
	g_debug("Active arrival! condition = 0x%x (NVAL = 0x%x, HUP = 0x%x, ERR = 0x%x)",condition,G_IO_NVAL,G_IO_HUP,G_IO_ERR);
	if(condition & G_IO_NVAL || condition & G_IO_HUP || condition & G_IO_ERR || hub == NULL) return FALSE;
	if(condition & G_IO_IN || condition & G_IO_PRI)
	{
		char * buf = g_malloc0(SR_PACKET_SIZE+1);
		g_debug("read_end = %d, buffer:",udp_socket_read(source,buf,SR_PACKET_SIZE));
		g_debug("\"%s\"",buf);
		notify_result(hub,buf);
		g_free(buf);
	}
	return TRUE;
}

/* annulla <n> ricerche per l'hub <hub> */
void dc_search_cancel(hub_t * hub,guint32 count)
{
#if 0
	guint32 * sid = NULL;
	char mode;
	g_debug("(dc_search_cancel): hub = %p, count = %d",hub,count);
	if(hub != NULL)
	{
		mode = userinfo_get_mode(hub->userinfo);
		sid = (mode == TAG_MODE_ACTIVE)?&search_act_count:&search_pas_count;
		g_debug("*sid (before) = %d for mode '%c'",*sid,mode);
		g_debug("count = %d, count for hub = %d",count,hub->search_count);
		if((*sid) <= 0) goto wait_thread;	// gia` tutto chiuso ;)
		if(count > hub->search_count)
		{
			(*sid) -= hub->search_count;
			hub->search_count = 0;
		}
		else
		{
			(*sid) -= count;
			hub->search_count -= count;
		}
		g_debug("*sid (after) = %d for mode '%c'",*sid,mode);
wait_thread:	// attendiamo il thread se *sid <= 0 (LOOL :)
		if((*sid) <= 0)
		{
			g_debug("(dc_search_cancel): waiting for %s thread",(mode==TAG_MODE_ACTIVE)?"active":"passive");
			GThread ** thread = (mode == TAG_MODE_ACTIVE)?&active_thread:&passive_thread;
			//g_debug("(dc_search_cancel): thread is %p",*thread);
			g_thread_join(*thread);
			*thread = NULL;
		}
	}
#endif
}

void destroy_search(search_t * se)
{
	g_free(se->host_nick);
	g_free(se->pattern);
	g_free(se);
}

static void destroy_internal_result(internal_result_t * res)
{
	g_free(res->sr_packet);
	g_free(res);
}

void destroy_result(result_t * r)
{
	g_free(r->nick);
	g_free(r->filename);
	g_free(r->extendedprotocol.hubname);
	g_free(r->extendedprotocol.tth);
	g_free(r->hubhost);
	g_free(r);
}

/****************************************************************
*								*
*			  UDP SOCKET				*
*								*
****************************************************************/

/* crea e binda una socket udp in ascolto su <ip>:<port>
 se ip == NULL binda su qualunque ip
 restituisce il GIOChannel associato a tale socket */
static GIOChannel * udp_socket_create(const char * ip,port_t port)
{
	/* converti l'ip in in_addr */
	struct sockaddr_in serv_addr;
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);
	if(ip != NULL)
	{
		struct in_addr ipaddr;
		//g_debug("atoning IP: \"%s\"",ip);
		if(!inet_aton(ip,&ipaddr))
		{
			g_warning("(udp_socket_create): invalid bind IP");
			goto error1;
		}
		memcpy((void *)&serv_addr.sin_addr,(void *)&ipaddr,sizeof(struct in_addr));
	}
	else serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	/* crea la socket */
	int sock_fd = socket(PF_INET,SOCK_DGRAM,0);
	int i = 1;
	if(setsockopt(sock_fd,SOL_SOCKET,SO_REUSEADDR,&i, sizeof(i)) < 0)
	{
		g_critical("(udp_socket_create): setsockopt() error");
		goto error2;
	}

	if(bind(sock_fd,(struct sockaddr *)&serv_addr,sizeof(struct sockaddr)) < 0)
	{
		g_critical("(udp_socket_create): Bind error (%s)",strerror(errno));
		goto error2;
	}
	GIOChannel * ioc = g_io_channel_unix_new(sock_fd);
	if(ioc != NULL) g_io_channel_set_encoding(ioc,NULL,NULL);
	return ioc;
error2:
	close(sock_fd);
error1:
	return FALSE;
}

/* legge dalla socket udp presa da <ioc>
 restituisce -1 in caso di errore, -2 in caso di risorsa non disponibile (nessun dato), 0 in caso di EOF */
static int udp_socket_read(GIOChannel *ioc,char * buf,size_t len)
{
	if(ioc == NULL || buf == NULL) return -1;
	if(!len) return -2;	// li leggi te zero byte :P
	int i,l=0,sock_fd = g_io_channel_unix_get_fd(ioc);
	socklen_t cli_len;		// UDP
	struct sockaddr cli_addr;	// UDP
	struct timeval tv;
	fd_set readset;
	cli_len = sizeof(cli_addr);	// UDP
	FD_ZERO(&readset);
	FD_SET(sock_fd,&readset);
	tv.tv_sec  = 0;
	tv.tv_usec = 0;
	i = select(FD_SETSIZE,&readset,NULL,NULL,&tv);
	if((i > 0) && (FD_ISSET(sock_fd,&readset)))
	{
		/* leggi dalla socket */
		l = recvfrom(sock_fd,buf,len,0,&cli_addr,&cli_len);	// UDP
		if(l < 0) if(errno == EINTR || errno == EAGAIN) return -2;
	}
	else if(i < 0)	// errore
	{
		g_debug("(dc_socket_read): read error");
		return -1;
	}
	else	// il tempo e` passato (nessun dato)
	{
		//g_debug("(dc_socket_read): read timeout");
		return -2;
	}
	if((i > 0) && (!l)) // siamo disconnessi
	{
		g_debug("(dc_socket_read): connection reset by peer");
		return 0;
	}
	FD_CLR(sock_fd,&readset);
	return l;
}
