/* userlist.c -- lista utenti */
/******************************************************************************
* DCLibC - Direct Connect C library                                           *
*                                                                             *
* Copyright (C) 2005-2007 daniele_athome                                      *
*                                                                             *
* This program is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU General Public License                 *
* as published by the Free Software Foundation; either version 2              *
* of the License, or (at your option) any later version.                      *
*                                                                             *
* This program is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               *
* GNU General Public License for more details.                                *
*                                                                             *
* You should have received a copy of the GNU General Public License           *
* along with this program; if not, write to the Free Software                 *
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. *
*******************************************************************************/

#define _GNU_SOURCE
#include <string.h>
#include <glib.h>
#include "userlist.h"
#include "userinfo.h"

static int userlist_get_index_from_nick(userlist_t * list,const char * nick);

/* crea una nuova userlist */
userlist_t * userlist_new(void)
{
	userlist_t * list = g_ptr_array_new();
	if(list == NULL) return NULL;
	return list;
}

/* distrugge una userlist */
void userlist_delete(userlist_t * list)
{
	if(list != NULL)
	{
		/* libera tutte le userinfo prima */
		g_ptr_array_foreach(list,(GFunc)userinfo_delete,NULL);
		g_ptr_array_free(list,TRUE);
		list = NULL;
	}
}

/* aggiunge un utente alla userlist
 restituisce FALSE in caso di errore o se l'utente esiste già, altrimenti TRUE*/
gboolean userlist_add(userlist_t * list,userinfo_t * user)
{
	if(list == NULL || user == NULL) return FALSE;
	/* guarda se l'utente esiste già */
	if(userlist_get_index_from_nick(list,user->nick) >= 0) return FALSE;
	/* aggiungi la userinfo alla userlist */
	g_debug("(userlist_add): adding user %s",user->nick);
	userinfo_t * u = userinfo_copy(user);
	//g_debug("(userlist_add): new user: %p",u);
	//g_debug("(userlist_add): old nick: %p, new nick: %p",user->nick,u->nick);
	g_ptr_array_add(list,u);
	return TRUE;
}

/* cerca e restituisce le userinfo per il nick <nick> nella lista <list>
 la userinfo restituita e` l'originale preso dalla userlist, quindi SI PREGA DI NON MODIFICARLA */
const userinfo_t * userlist_get_from_nick(userlist_t * list,const char * nick)
{
	//g_debug("(userlist_get_from_nick): nick = \"%s\"",nick);
	int ind = userlist_get_index_from_nick(list,nick);
	//g_debug("(userlist_get_from_nick): ind = %d",ind);
	if(ind >= 0) return g_ptr_array_index(list,ind);
	return NULL;
}

/* svuota la userlist */
void userlist_clear(userlist_t * list)
{
	if(list != NULL)
	{
		if(list->len > 0)
		{
			g_ptr_array_foreach(list,(GFunc)userinfo_delete,NULL);
			g_ptr_array_remove_range(list,0,list->len);
		}
	}
}

/* rimuove il nick dalla lista
 restituisce FALSE se non e` stato trovato in lista */
gboolean userlist_remove(userlist_t * list,const char * nick)
{
	if(list == NULL || nick == NULL) return FALSE;
	int ind = userlist_get_index_from_nick(list,nick);
	if(ind < 0) return FALSE;
	/* rimuovi dalla ptrarray */
	userinfo_t * u = g_ptr_array_index(list,ind);
	g_debug("(userlist_remove): removing user %s",u->nick);
	g_ptr_array_remove_index(list,ind);
	userinfo_delete(u);
	return TRUE;
}

/* cerca <nick> nella lista, se lo trova mergia la entry con <user>
 se non lo trova lo aggiunge con <user> */
void userlist_update(userlist_t * list,userinfo_t * user)
{
	if(list == NULL || user == NULL) return;
	userinfo_t * u = NULL;
	//g_debug("(userlist_update): looking for nick %s",user->nick);
	int ind = userlist_get_index_from_nick(list,user->nick);
	if(ind < 0)	// non trovato, aggiungi
	{
		g_debug("(userlist_update): nick %s not found, adding...",user->nick);
		userlist_add(list,user);
	}
	else	// trovato, aggiorna
	{
		u = g_ptr_array_index(list,ind);
		g_debug("(userlist_update): nick %s (%p) found, merging... (u: \"%s\", %p)",user->nick,user,u->nick,u);
		userinfo_t * old = g_ptr_array_index(list,ind);
		g_ptr_array_index(list,ind) = userinfo_merge(u,user);
		userinfo_delete(old);
	}
}

static int userlist_get_index_from_nick(userlist_t * list,const char * nick)
{
	if(list == NULL || nick == NULL) return -1;
	userinfo_t * u = NULL;
	unsigned int i;
	for(i=0;i<list->len;i++)
	{
		u = g_ptr_array_index(list,i);
		//if(u != NULL) g_debug("(get_index_from_nick): u->nick = \"%s\", nick = \"%s\"",u->nick,nick);
		if(u != NULL) if(!g_ascii_strcasecmp(u->nick,nick)) return i; // corrisponde!!
	}
	return -1;
}
